from collections.abc import Iterable

import dcase_util as dcu
import sed_eval


# TODO change eb to event base somethin
def eb_evaluator(y_true, y_pred,
              t_collar: float = 0.200,
              percentage_of_length: float = 0.2
              ) -> sed_eval.sound_event.EventBasedMetrics:
    # TODO In the sed_eval library, EventBasedMetrics have more attribute.
    """
    Return an evaluator function depending on the type of the data provided.
    Different type of data are possible to use. Sed_eval event based metrics.

    # Exemple
    # A list of string
    estimated_event_list = [
        ["file_1.wav\t1.00\t2.00\dog"],
        ["file_2.wav\t1.56\t1.92\cat"]
    ]

    # A string
    estimated_event_list = "file_1.wav\t1.00\t2.00\dog\n" + \
        "file_2.wav\t1.56\t1.92\cat"

    Args:
        y_true: The ground truth
        y_pred: The prediction (results of the encode function)
        t_collar (float): Time collar used when evaluating validity of the onset
        and offset, in seconds. Default value 0.2
        percentage_of_length (float): Second condition, percentage of the
        length within which the estimated offset has to be in order to be
        consider valid estimation. Default value 0.2

    Returns:
        EventBasedMetrics: **event_based_metric**
    """

    # Convert the data into dcase_util.containers.MetaDataContainer
    _y_true = convert_to_mdc(y_true)
    _y_pred = convert_to_mdc(y_pred)

    # return _y_true.unique_files, _y_pred.unique_files
    return event_based_evaluation(_y_true, _y_pred,
                                  t_collar, percentage_of_length)

    # print(_y_true.unique_event_labels)
    # event_based_metric = sed_eval.sound_event.EventBasedMetrics(
    #     event_label_list=_y_true.unique_event_labels,
    #     t_collar=t_collar,
    #     percentage_of_length=percentage_of_length,
    # )

    # return event_based_metric.evaluate(
    #     _y_true, _y_pred
    # )

def event_based_evaluation(reference_event_list, estimated_event_list,
                           t_collar: float, percentage_of_length: float):
    """Calculate sed_eval event based metric for challenge

    Args:
        reference_event_list (MetaDataContainer, list of referenced events):
        estimated_event_list (MetaDataContainer, list of estimated events):

    Returns:
        EventBasedMetrics: **event_based_metric**
    """

    files = {}
    for event in reference_event_list:
        files[event['filename']] = event['filename']

    evaluated_files = sorted(list(files.keys()))

    event_based_metric = sed_eval.sound_event.EventBasedMetrics(
        event_label_list=reference_event_list.unique_event_labels,
        t_collar=t_collar,
        percentage_of_length=percentage_of_length,
    )

    for file in evaluated_files:
        reference_event_list_for_current_file = []
        # events = []
        
        for event in reference_event_list:
            if event['filename'] == file:
                reference_event_list_for_current_file.append(event)
                # events.append(event.event_label)

        estimated_event_list_for_current_file = []
        for event in estimated_event_list:
            if event['filename'] == file:
                estimated_event_list_for_current_file.append(event)

        # Replace
        # replace every None by ""
        def none2empty(root):
            for d in root:
                for k in d:
                    if d["event_label"] is None:
                        d["event_label"] = "Speech"

        # none2empty(reference_event_list_for_current_file)
        # none2empty(estimated_event_list_for_current_file)

        event_based_metric.evaluate(
            reference_event_list=reference_event_list_for_current_file,
            estimated_event_list=estimated_event_list_for_current_file
        )

    return event_based_metric


def sb_evaluator(y_true, y_pred,
                 time_resolution) -> sed_eval.sound_event.SegmentBasedMetrics:

    # TODO add doc
    # convert the data into dcase_util.containers.MetadataContainer
    _y_true = convert_to_mdc(y_true)
    _y_pred = convert_to_mdc(y_pred)

    return segment_based_evaluation(_y_true, _y_pred, time_resolution)

def segment_based_evaluation(reference_event_list, estimated_event_list,
                             time_resolution):
    # TODO add documentation
    segment_based_metrics = sed_eval.sound_event.SegmentBasedMetrics(
        event_label_list=reference_event_list.unique_event_labels,
        time_resolution=time_resolution
    )

    for filename in reference_event_list.unique_files:
        reference_event_list_for_current_file = reference_event_list.filter(
            filename=filename
        )

        estimated_event_list_for_current_file = estimated_event_list.filter(
            filename=filename
        )

        segment_based_metrics.evaluate(
            reference_event_list=reference_event_list_for_current_file,
            estimated_event_list=estimated_event_list_for_current_file
        )

    return segment_based_metrics

# ==============================================================================
#
#       CONVERTION FUNCTIONS
#
# ==============================================================================
def __detect_separator(exemple: str) -> str:
    """
    Automatically detect the separator use into a string and return it

    Args:
        A String exemple

    Returns:
        separator character
    """
    known_sep = [",", ";", ":", "\t"]

    for sep in known_sep:
        if len(exemple.split(sep)) > 1:
            return sep

    return "\t"


def convert_to_mdc(event_list) -> dcu.containers.MetaDataContainer:
    """
    Since the user can provide the reference and estimates event list in
    different format, We must convert them into MetaDataContainer.

    Args:
        event_list: The event list into one of the possible format

    Returns:
        MetaDataContainer
    """

    if isinstance(event_list, dcu.containers.MetaDataContainer):
        return event_list

    # A string
    elif isinstance(event_list, str):
        return string_to_mdc(event_list)

    # list of string
    elif isinstance(event_list, Iterable):
        if isinstance(event_list[0], str):
            return list_string_to_mdc(event_list)

    else:
        raise ValueError("This format %s can't be used. " % type(event_list))


def string_to_mdc(event_strings: str) -> dcu.containers.MetaDataContainer:
    """
    If the data is under the form of a long string with several line (\n).
    The information contain in each line must be separated using one of this
    separator : ",", ";", "\t". It will be automatically detected.

    Args:
         event_strings (str): The string to convert into a MetaDataContainer


    Returns:
        MetaDataContainer
    """
    list_json = []

    # Automatically find the separator
    sep = __detect_separator(event_strings.split("\n")[0])

    for line in event_strings.split("\n"):

        # TODO automatic separator
        info = line.split("\t")

        if len(info) != 4:
            continue

        info_dict = dict()
        info_dict["file"] = info[0]
        if info[1] != "": info_dict["event_onset"] = float(info[1])
        if info[2] != "": info_dict["event_offset"] = float(info[2])
        if info[3] != "": info_dict["event_label"] = info[3]

        list_json.append(info_dict)

    return dcu.containers.MetaDataContainer(list_json)


def list_string_to_mdc(event_list: list) -> dcu.containers.MetaDataContainer:
    """
    If the data is under the form of a list of strings. The information contain
    in each line must be separated using one of this separator : ",", ";",
    "\t". It will be automatically detected.

    Args:
         event_list (str): The string to convert into a MetaDataContainer


    Returns:
        MetaDataContainer
    """
    list_json = []

    # Automatically find the separator
    sep = __detect_separator(event_list[0])

    for line in event_list:
        info = line.split("\t")

        # TODO automatic separator
        info = line.split("\t")

        if len(info) != 4:
            continue

        info_dict = dict()
        info_dict["file"] = info[0]
        if info[1] != "": info_dict["event_onset"] = float(info[1])
        if info[2] != "": info_dict["event_offset"] = float(info[2])
        if info[3] != "": info_dict["event_label"] = info[3]

        list_json.append(info_dict)

    return dcu.containers.MetaDataContainer(list_json)


if __name__=='__main__':
    import numpy as np
    from aeseg.Encoder import Encoder

    # load baseline data
    strong_prediction_path = "/home/lcances/sync/Documents_sync/Projet" \
                        "/Threshold_optimization/data/baseline_strong_prediction.npy"
    strong_prediction = np.load(strong_prediction_path)

    # classes
    class_correspondance = {"Alarm_bell_ringing": 0, "Speech": 1, "Dog": 2,
                            "Cat": 3, "Vacuum_cleaner": 4,
                            "Dishes": 5, "Frying": 6,
                            "Electric_shaver_toothbrush": 7, "Blender": 8,
                            "Running_water": 9}
    class_list = list(class_correspondance.keys())

    encoder = Encoder(class_list, 200, 10, 200)

    import time
    def test(name):
        start = time.time()
        segments = encoder.encode(strong_prediction, method=name)
        end = time.time() - start


    test("threshold")
    test("hysteresis")
    test("derivative")
    test("mean_threshold")
    test("median_threshold")
    test("global_mean_threshold")
    test("global_median_threshold")



import functools
import itertools
import operator
from collections.abc import Iterable
from multiprocessing import Pool

import numpy as np
import tqdm

from Encoder import Encoder


class Optimizer:
    def __init__(self, step: int, nb_recurse: int, nb_process: int = 4):
        """
        Abstract class for the different optimizer available

        .. seealso::
            - :class:`optimizers.DichotomicOptimizer`
            - :class:`optimizers.GenOptimizer`

        Args:
            param (dict):
                The parameters that should be optimized.
            optimize_fn (Callable):
                The encoder object that will be used to compute the score
            step (int):
                In how many step the range should be divided.
            nb_recurse (int):
                The number of recursion.
            nb_process (int):
                The number of processes used for the optimization.
        """

        self.step = step
        self.nb_recurse = nb_recurse

        # Synchronization variables
        self.nb_process = nb_process
        self.process_pool = None
        self.results = dict()
        self.fitted = False

        # Monitoring
        self.progress = None

    def find(self, monitor: tuple, data: dict):
        """ Find the metric to monitor in the nested dictionary return by the evaluator.
        """
        return functools.reduce(operator.getitem, monitor, data)

    def nb_iteration(self, parameters):
        raise NotImplementedError

    def dict_nan_to_num(self, d: dict):
        """ For every np.nan within the dictionary and the nested dictionary, will convert every nan to 0.0.
        """
        for k in d:
            if isinstance(d[k], dict):
                self.dict_nan_to_num(d[k])
            else:
                d[k] = np.nan_to_num(d[k])

    def fit(self, kwds=dict(), verbose=1):
        """ Perform the optimization of the parameters provided by kwds.
        kwds (dict):
            The dictionnary key represents the parameters of the optimization function
            Each key value must be either a list, a tuple or a single value.
            A tuple will be interpreted as a range of value possible, automatially split into <steps>
            A list will be interpreted as an ensemble fix of value to test
            A single value as a fixed value (no optimization on this parameter
        """
        self.fitted = True
        self.keys = list(kwds.keys())

        # start the pool
        self.process_pool = Pool(self.nb_process)
        
        if verbose == 1:
            self.progress = tqdm.tqdm(total=self.nb_iteration(kwds))
        if verbose == 2:
            self.progress = tqdm.tqdm_notebook(total=self.nb_iteration(kwds))

    @property
    def history(self) -> dict:
        """
        Return the complete history of the optimization process, useful for visualization of the optimization process.
        Work only if an optimization process have been already done.

        Returns:
            a list of dict, see sed_eval documentation for further detail
        """
        if not self.fitted:
            raise RuntimeWarning("No optimization done yet")
        return self.results

    @property
    def best(self) -> tuple:
        """
        Return the combination of parameters that yield the best score using the monitored metric.

        Returns:
            tuple (parameters, score)
        """
        if not self.fitted:
            raise RuntimeWarning("Optimization not perform yet")

        out = []

        for k in self.results.keys():
            value = self.results[k]
            out.append((k, value))

        out = sorted(out, key=lambda x: x[1])

        # pair each value with it parameter
        return dict(zip(self.keys, out[-1][0])), out[-1][1]


class DichotomicOptimizer(Optimizer):
    """ The dichotomous optimizer aims to drastically reduce the search time for the combination of parameters
    that gives the best score.

    For each parameter that needs to be optimized, the user provides a search interval. Using a "step" parameter
    defined at creation, the optimizer will then search for all possible combinations. The best combination is
    selected, and new intervals are calculated. The optimizer then resumes its search in these new intervals.
    The optimization stops when a fixed number of recursions specified at creation is reached.
    """
    def __init__(self, step: int,
                 nb_recurse: int, nb_process: int = 4):
        """
        Args:
            param (dict):
                The parameters that should be optimized.
            optimize_fn (Callable):
                The function used to compute the metric to optimize
            step (int):
                In how many step the range should be divided.
            nb_recurse (int):
                The number of recursion.
            nb_process (int):
                The number of processes used for the optimization.

        :Exemple:
        To use the Dichotomous optimizer on the absolute threshold method, 3 step are needed.
        - Define the list of class that are predicted by the system.
        - Create an encoder with a specified temporal prediction and a minimum segmentatio separation.
        - Create an optimizer that will use this encoder and perform best combination search.

        .. code-block:: python

            class_list = ['Alarm_bell_ringing', 'Speech', 'Dog', 'Cat', 'Vacuum_cleaner', 'Dishes', 'Frying', 'Electric_shaver_toothbrush', 'Blender', 'Running_water']

            # Create the encoder that will be used
            encoder = Encoder(
                classes=class_list,
                temporal_precision = 200,  # ms
                clip_length = 10,          # s
                minimal_segment_step = 200 # ms
            )

            parameter_to_optimize = {
                    "threshold": (0.1, 0.9),
                    "smooth": "smoothMovingAvg",
                    "window_len": (5, 27)
                }

            optimizer = DichotomicOptimizer(
                **parameter_to_optimize,
                method="threshold",
                encoder = encoder,

                step = 6,
                nb_recurse = 10,
                nb_process = 20
            )
        """
        super().__init__(step, nb_recurse, nb_process)

    def param_to_range(self, param: dict) -> dict:
        """Giving a tuple representing the minimum and the maximum value of a
        parameters, will generate, uniformly a list of value to test.

        Args:
            param (dict): The parameters to optimize
        """
        outputs = dict()

        for key in param:
            # is tuple --> range
            if isinstance(param[key], tuple):
                outputs[key] = np.linspace(
                    param[key][0], param[key][1],
                    self.step
                )

            # if list --> actual liste
            elif isinstance(param[key], list):
                outputs[key] = param[key]

            # if not iterable --> fix value, no changing
            elif not isinstance(param[key], Iterable):
                outputs[key] = param[key]

            # if str --> fix value, no changing
            elif isinstance(param[key], str):
                outputs[key] = [param[key]]

        return outputs

    def nb_iteration(self, parameters):
        extend_parameters = self.param_to_range(parameters)

        nb_iteration = 1
        for key in extend_parameters:
            nb_iteration *= len(extend_parameters[key])

        return nb_iteration * self.nb_recurse

    def two_best(self, source: dict, params: dict, nb_recurse: int = 1, mode="max")-> dict:
        """Return the new range (tuples) for each parameters based on the two
        best results

        Args:
            source (dict): The combination of parameters, the keys are a tuple
            keys (list): The list of combination
        """
        # Transform the dictionary into a list of tuple where the first element
        # is the combination of parameters and the second the score
        tuples = list(zip(source.keys(), source.values()))

        # Sort the combination by the score
        tuples.sort(key=lambda elem: elem[1])

        # best parameters combination
        pick = -1 if mode == "max" else 0
        best_combination = dict(zip(params.keys(), tuples[-1][0]))

        # For each parameters that need it (tuple), build the new search
        # boundaries
        for key in best_combination:
            # only for tuple range
            if isinstance(params[key], tuple):
                best = best_combination[key]
                half = best / (1 + nb_recurse)

                # boundaries
                low = best - half
                high = best + half

                # clip
                low = low if low > params[key][0] else params[key][0]
                high = high if high < params[key][1] else params[key][1]

                best_combination[key] = (low, high)

        # in some case, like with str param or unique value param,
        # the tuple created is no suitable, so we take only the first element
        # to go back to a unique parameter
        for key in best_combination:
            # if suppose to be str --> str
            if isinstance(params[key], str):
                best_combination[key] = best_combination[key]

            if isinstance(params[key], list):
                best_combination[key] = params[key]

        return best_combination

    def fit(self, optimize_fn, kwds=dict(), mode="max", verbose=1):
        super().fit(kwds, verbose)
        _param = kwds.copy()
        _args = kwds.keys()

        for recurse in range(self.nb_recurse):

            # Create all the combination
            search_space = self.param_to_range(_param)
            all_combination = itertools.product(*list(search_space.values()))

            # Add all combination to the thread pool
            works = []
            for combination in all_combination:
                fn_args = dict(zip(_args, combination))
                works.append((
                    combination,
                    self.process_pool.apply_async(optimize_fn, kwds=fn_args)
                ))

            # wait for the pool to finish working on the current recursion
            # Save all results in the history. The combination will be the key
            for combination, res in works:
                self.results[combination] = res.get()

                if verbose != 0:
                    self.progress.update()

            # Find the two best results among this recursion
            two_best = self.two_best(
                self.results,
                _param,
                recurse,
                mode
            )

            # Set the new parameters range for the next recursion
            _param = two_best

        # close the pool
        self.process_pool.close()
        self.process_pool.join()



from collections.abc import Iterable

import numpy as np


class Encoder:
    """ Allow the localization in time of sound events using temporal prediction.

    In a sound event detection task, the output of the prediction model is
    often a temporal prediction. Different segmentation algorithm exist in order
    to translate this curves into a list of segment. This encoder give you access
    to most of them in a fairly easy way.

    """

    def __init__(self, classes: list, temporal_precision: int, clip_length: int,
                 minimal_segment_step: int, t_collar: float = 0.200,
                 percentage_of_length: float = 0.2,
                 time_resolution: float = 1.00,
                 method: str = "event_based_metrics"):
        """ Initialization of the encoder.

        To initialize the encoder, you must provide the list of the classes that each curve will represent, \
        in the same order along with information about the curves and the precision of the segmentation.

        Args:
            classes (list):
                The list of class that each curves will represent It is require for the function parse.
            temporal_precision (int):
                The temporal prediction for each classes.
            clip_length (int):
                The length of the audio file (in seconds).
            minimal_segment_step (int):
                The minimum space in between two segments.

        :Exemple:

        ::

            class_list = ['Alarm_bell_ringing', 'Speech', 'Dog', 'Cat', 'Vacuum_cleaner', 'Dishes', 'Frying', 'Electric_shaver_toothbrush', 'Blender', 'Running_water']

            # Create the encoder that will be used
            encoder = Encoder(
                classes=class_list,
                temporal_precision = 200,  # ms
                clip_length = 10,          # s
                minimal_segment_step = 200 # ms
            )
        """
        self.classes = classes
        self.temporal_precision = temporal_precision
        self.clip_length = clip_length
        self.minimal_segment_step = minimal_segment_step
        self.t_collar = t_collar
        self.percentage_of_length = percentage_of_length
        self.time_resolution = time_resolution
        self.method = method

        # Attribute that are not initialize with the constructor
        self.frame_length = None
        self.nb_frame = None
        self.class_correspondence = dict(zip(classes, range(len(classes))))
        self.class_correspondence_reverse = dict(zip(range(len(classes)), classes))

    def encode(self, temporal_prediction: np.array, method: str = "threshold",
               smooth: str = None, **kwargs) -> list:
        """Perform the localization of the sound event present in the file.

        Using the temporal prediction provided y the last step of the system, it will "localize" the sound event
        inside the file under the form of a strongly annotated line. (see DCASE2018 task 4 strong label exemple).
        There is two methods implemented here, one using a simple threshold based segmentation and an other using
        a modulation system based on the variance of the prediction over the time.

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy array
                (<sample>, <class>, <frames>)
            method (str):
                The segmentation method to use [threshold | hysteresis | derivative | primitive |
                mean_threshold | global_mean_threshold | median_threshold | gobal_median_threshold].
            smooth (str):
                The smoothing method to use [smoothMovingAvg].
            kwargs:
                See the segmentation method parameters.

        Returns:
            Return a list of positive and negative segments with their size. A
            segment is a tuple where the first value that represent the segment
            value (1) for positive, (0) for negative and the second values is
            the width of the segment (number of frame).
        """
        # parameters verification
        _methods = ["threshold", "hysteresis", "derivative", "mean_threshold", "median_threshold", "dynamic_threshold",
                    "global_mean_threshold", "global_median_threshold"]

        if method not in _methods:
            raise ValueError("Method %s doesn't exist. Only %s are available" % (method, _methods))

        # Depending on the method selected, the proper function will be selected
        encoder = None

        if method == _methods[0]:
            encoder = self._encode_using_threshold
        elif method == _methods[2]:
            encoder = self._encode_using_derivative
        elif method == _methods[1]:
            encoder = self._encode_using_hysteresis
        elif method == _methods[3]:
            encoder = self._encode_using_mean_threshold
        elif method == _methods[6]:
            encoder = self._encode_using_gmean_threshold
        elif method == _methods[4]:
            encoder = self._encode_using_median_treshold
        elif method == _methods[5]:
            encoder = self._encode_using_dynamic_threshold
        elif method == _methods[7]:
            encoder = self._encode_using_gmedian_threshold

        # Apply smoothing if requested
        if smooth is not None:
            temporal_prediction = self._smooth(temporal_prediction,
                                               method=smooth, **kwargs)

        # Now that we have the strong prediction, we can assign the value to the
        # two attributes nb_frame and frame_length
        self.nb_frame = temporal_prediction.shape[1]
        self.frame_length = self.clip_length / self.nb_frame

        # Execute the selected segmentation algorithm and recover its results
        return encoder(temporal_prediction, **kwargs)

    def parse(self, all_segments: list, test_files_name: list) -> str:
        """ Transform a list of segment into a string ready for evaluation with sed_eval.

        Args:
            all_segments (list): a list of dict of 10 key. the list length is equal to the number of file, the dict \
                number test_files_name.
            test_files_name(list): The list of the file names in the same.
        """
        output = ""

        for clipIndex in range(len(all_segments)):
            clip = all_segments[clipIndex]

            empty_cls = 0
            for cls in clip:
                if len(clip[cls]) == 1 and clip[cls][0][0] == 0:
                    empty_cls += 1

            if empty_cls == 10:
                output += "%s\n" % test_files_name[clipIndex]
            else:

                for cls in clip:
                    start = 0
                    for segment in clip[cls]:
                        if segment[0] == 1.0:
                            output += "%s\t%f\t%f\t%s\n" % (
                                test_files_name[clipIndex],
                                start * self.frame_length,
                                (start + segment[1]) * self.frame_length,
                                self.class_correspondence_reverse[cls]
                            )
                        start += segment[1]

        return output

    # ==================================================================================================================
    #
    #       ENCODING METHODS
    #
    # ==================================================================================================================
    def _encode_using_threshold(self, temporal_prediction: np.array, threshold: float or list, **kwargs) -> list:
        """A basic threshold segmentation algorithm.

        For each frame where the probability is above the given threshold, will be part of a valid segment, \
        an invalid one otherwise. The threshold can be set globally (one unique threshold for all the \
        classes) or independently (one threshold for each classes)

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy array
                (<sample>, <class>, <frames>)
            threshold (float or list):
                One unique threshold or a list of threhsold. If using a list, it must define one threshold for
                each class.
            kwargs:
                kwargs ...

        :Exemple:

        ::

            segments = encoder.encode(
                val_strong_pred,
                method="threshold",
                threshold=0.5,
               **smoothing_parameters  # (optional)
            )
        """
        output = []

        # Treshold can be either an int or an Iterable (list, np.Array, tuple, etc ...)
        # If int, transform it to a list
        thresholds = threshold
        if not isinstance(threshold, Iterable):
            thresholds = [threshold] * len(self.classes)

        bin_prediction = temporal_prediction.copy()
        bin_prediction[bin_prediction > thresholds] = 1
        bin_prediction[bin_prediction <= thresholds] = 0

        # Merging "hole" that are smaller than 200 ms
        step_length = self.clip_length / temporal_prediction.shape[1] * 1000
        max_hole_size = int(self.minimal_segment_step / step_length)

        for clip in bin_prediction:
            labeled = dict()

            cls = 0
            for bin_prediction_per_class in clip.T:
                # convert the binarized list into a list of tuple representing
                # the element and it's number of occurrence. The order is
                # conserved and the total sum should be equal to 10s

                # first pass --> Fill the holes
                for i in range(len(bin_prediction_per_class) - max_hole_size):
                    window = bin_prediction_per_class[i: i + max_hole_size]

                    if window[0] == window[-1] == 1:
                        window[:] = [window[0]] * max_hole_size

                # second pass --> split into segments
                converted = []
                cpt = 0
                nb_segment = 0
                previous_elt = None
                for element in bin_prediction_per_class:
                    if previous_elt is None:
                        previous_elt = element
                        cpt += 1
                        nb_segment = 1
                        continue

                    if element == previous_elt:
                        cpt += 1

                    else:
                        converted.append((previous_elt, cpt))
                        previous_elt = element
                        nb_segment += 1
                        cpt = 1

                # case where the class is detect during the whole clip
                #                 if nbSegment == 1:
                converted.append((previous_elt, cpt))

                labeled[cls] = converted.copy()
                cls += 1

            output.append(labeled)

        return output

    def _encode_using_hysteresis(self, temporal_prediction: np.array, low: float = 0.4, high: float = 0.6, **kwargs) -> list:
        """ Hysteresis threshold segmentation method.

        The hysteresis based segmentation algorithm require two thresholds. A high value to decided when the
        segment should start and a low value to decided when to finish the segment. It perform better when the temporal
        prediction is noisy.

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy
                array (<sample>, <class>, <frames>).
            low (float):
                low threshold (can be a list for class-dependant thresholding)
            high (float):
                high threshold (can ve a list for class-dependant thresholding)
            kwargs:
                Extra arguments

        Returns:
            the result of the system under the form of a strong annotation text where each line represent on timed event.

        :Exemple:

        ::

            segments = encoder.encode(
                val_strong_pred,
                method="hysteresis",
                low=0.4,
                high=0.6
               **smoothing_parameters  # (optional)
            )

        """

        # In case of class dependant thresholding
        lows = low
        if not isinstance(low, Iterable):
            lows = [low] * len(self.classes)

        highs = high
        if not isinstance(high, Iterable):
            highs = [high] * len(self.classes)

        prediction = temporal_prediction

        output = []

        for clip in prediction:
            labeled = dict()

            cls = 0
            for cls_ind, prediction_per_class in enumerate(clip.T):
                converted = list()
                segment = [0, 0]
                nb_segment = 1
                for i in range(len(prediction_per_class)):
                    element = prediction_per_class[i]

                    # first element
                    if i == 0:
                        if element > highs[cls_ind]:
                            segment = [1.0, 1]
                        else:
                            segment = [0.0, 1]

                    # then
                    if element > highs[cls_ind]:
                        if segment[0] == 1:
                            segment[1] += 1
                        else:
                            converted.append(segment)
                            nb_segment += 1
                            segment = [1.0, 1]

                    elif lows[cls_ind] <= element:
                        segment[1] += 1

                    else:
                        if segment[0] == 0:
                            segment[1] += 1
                        else:
                            converted.append(segment)
                            nb_segment += 1
                            segment = [0.0, 0]

                converted.append(segment)

                labeled[cls] = converted.copy()
                cls += 1

            output.append(labeled)

        return output

    def _encode_using_derivative(self, temporal_prediction: np.array, rising: float = 0.5, decreasing: float = -0.5,
                                 window_size: int = 5, high: float = 0.8, padding: str = "same", **kwargs) -> list:
        """ Slope based segmentation.

        The derivative create segment based on the intensity of the variation of the temporal prediction curve.
        If the prediction rise above a certain threshold `rising` then a valid segment start. If it decrease faster
        than the `decreasing` threshold, then a valid segment finish. If the prediction start with a high value,
        of rise slowly but high, then an absolute (and global) threshold `high` is used. (it works like a normal
        threhsold).

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy
                array (<sample>, <class>, <frames>).
            rising (float):
                Must be between 0 and 1, rising threshold. When the decreasing (float): Must be between 0 and 1,
                decreasing threshold.
            window_size (int):
                size of the processing window.
            high (float):
                minimum prediction value that trigger a valid.
            padding (str):
                The padding method to used on the curves.

        Returns:
            The result of the system under the form of a strong annotation text where each represent on timed event.

        :Exemple:

        ::

            segments = encoder.encode(
                val_strong_pred,
                method="derivative",
                rising=0.5,
                decreasing=-0.5,
                window_size=5,
                high=0.9
                padding="same",
                **smoothing_parameters  # (optional)
            )

        """

        output = []

        # for class-dependant parameters
        window_sizes = window_size
        if not isinstance(window_size, Iterable):
            window_sizes = [window_size] * len(self.classes)

        risings = rising
        if not isinstance(rising, Iterable):
            risings = [rising] * len(self.classes)

        decreasings = decreasing
        if not isinstance(decreasing, Iterable):
            decreasings = [decreasing] * len(self.classes)

        highs = high
        if not isinstance(high, Iterable):
            highs = [high] * len(self.classes)


        for clip in temporal_prediction:
            cls = 0
            labeled = dict()

            for cls_ind, prediction_per_class in enumerate(clip.T):
                # get class-dependant parameters
                _window_size = int(window_sizes[cls_ind])
                _rising = risings[cls_ind]
                _decreasing = decreasings[cls_ind]
                _high = highs[cls_ind]

                padded_prediction_per_class = self._pad(prediction_per_class,
                                                        _window_size,
                                                        method=padding)

                nb_segment = 1
                segments = []
                segment = [0.0, 0]
                for i in range(len(padded_prediction_per_class) - _window_size):
                    window = padded_prediction_per_class[i:i + _window_size]
                    slope = (window[-1] - window[0]) / _window_size

                    # first element
                    if i == 0:
                        if window[0] > _high:
                            segment = [1.0, 1]
                        else:
                            segment = [0.0, 1]

                    # if on "high" segment
                    if segment[0] == 1:

                        # if above high threshol
                        if window[0] > _high:
                            segment[1] += 1

                        else:
                            # if decreasing threshold is reach
                            if slope < _decreasing:
                                segments.append(segment)
                                nb_segment += 1
                                segment = [0.0, 1]
                            else:
                                segment[1] += 1

                    # if on "low" segment
                    else:

                        # if above high threshold
                        if window[0] > _high:
                            segments.append(segment)
                            nb_segment += 1
                            segment = [1.0, 1]

                        else:
                            if slope > _rising:
                                segments.append(segment)
                                nb_segment += 1
                                segment = [1.0, 1]
                            else:
                                segment[1] += 1

                segments.append(segment.copy())

                labeled[cls] = segments
                cls += 1

            output.append(labeled)
        return output

    def _encode_using_gmean_threshold(self, temporal_prediction: np.array,
                                      independent: bool = False, **kwargs
                                      ) -> list:
        """ Absolute threshold computed using the average prediction of the whole dataset.

        Using all the temporal prediction, the mean of each curve and for each class is computed and will be choose
        as threshold. Then call the `__encode_using_threshold` function to apply it. The average is calculated
        independently for each class. Two return mode exist, one that return the average of class's threshold, giving
        an unique threshold for all the class, one that the return the list of threshold, giving a different threshold
        that depend on the class.

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy
                array (<sample>, <class>, <frames>)
            independent (bool):
                If True, return a list of threshold, one different for each class.
                If False, return an unique threshold
            kwargs:
                kwargs ...

        :Exemple:

        ::

            segments = encoder.encode(
                val_strong_pred,
                method="global_mean_threshold",
                **smoothing_parameters  # (optional)
            )

        """

        total_thresholds = []

        for clip in temporal_prediction:
            total_thresholds.append([curve.mean() for curve in clip.T])

        total_thresholds = np.array(total_thresholds)

        if independent:
            return self._encode_using_threshold(
                temporal_prediction,
                threshold=total_thresholds.mean(axis=0),
                **kwargs)
        else:
            return self._encode_using_threshold(
                temporal_prediction,
                threshold=[total_thresholds.mean()] * len(self.classes),
                **kwargs)

    def _encode_using_gmedian_threshold(self, temporal_prediction: np.array,
                                        independent: bool = False, **kwargs
                                        ) -> list:
        """ Absolute threshold computed using the median prediction of the whole dataset.

        Using all the temporal prediction, the median of each curve and for each class is selected and will be choose
        as threshold. Then call the `__encode_using_threshold` function to apply it. The median is retrieve
        independently for each class. Two return mode exist, one that return the average of class's threshold, giving
        an unique threshold for all the class, one that the return the list of threshold, giving a different threshold
        that depend on the class.

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy
                array (<sample>, <class>, <frames>)
            independent (bool):
                If True, return a list of threshold, one different for each class.
                If False, return an unique threshold
            kwargs:
                kwargs ...

        :Exemple:

        ::

            segments = encoder.encode(
                val_strong_pred,
                method="global_median_threshold",
                **smoothing_parameters  # (optional)
            )

        """

        total_thresholds = []
        for clip in temporal_prediction:
            # compute unique threshold for this file
            total_thresholds.append(
                [curve[len(curve) // 2] for curve in clip.T])

        total_thresholds = np.array(total_thresholds)

        if independent:
            return self._encode_using_threshold(
                temporal_prediction,
                threshold=total_thresholds.mean(axis=0),
                **kwargs)
        else:
            return self._encode_using_threshold(
                temporal_prediction,
                threshold=[total_thresholds.mean()] * len(self.classes),
                **kwargs)

    def _encode_using_mean_threshold(self, temporal_prediction: np.array, **kwargs) -> list:
        """ Absolute threshold computed using the average prediction **independently** for each file in the dataset

        Using all the temporal prediction from a specified file, the average of each curve is selected and chosen
        as threshold. Then call the `__encode_using_threshold` function to apply it. The average is retrieve
        independently for each class. Two return mode exist, one that return the average of class's threshold, giving
        an unique threshold for all the class, one that the return the list of threshold, giving a different threshold
        that depend on the class.

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy
                array (<sample>, <class>, <frames>)
            independent (bool):
                If True, return a list of threshold, one different for each class.
                If False, return an unique threshold
            kwargs:
                kwargs ...

        :Exemple:

        ::

            segments = encoder.encode(
                val_strong_pred,
                method="mean_threshold",
                **smoothing_parameters  # (optional)
            )

        """

        # Recover the kwargs arguments
        _global = kwargs.get("global", False)

        output = []

        # Merging "hole" that are smaller than 200 ms
        step_length = self.clip_length / temporal_prediction.shape[1] * 1000
        max_hole_size = int(self.temporal_precision / step_length)

        for clip in temporal_prediction:
            labeled = dict()
            _clip = clip.copy()

            # compute unique threshold for this file globally or independent
            thresholds = np.array([curve.mean() for curve in _clip.T])

            if _global:
                thresholds = [thresholds.mean()] * len(self.classes)

            # Binarize using the given thresholds
            if thresholds is not None:
                _clip[_clip > thresholds] = 1
                _clip[_clip <= thresholds] = 0

            cls = 0
            for binPredictionPerClass in _clip.T:
                # convert the binarized list into a list of tuple representing
                # the element and it's number of # occurrence. The order is
                # conserved and the total sum should be equal to 10s

                # first pass --> Fill the holes
                for i in range(len(binPredictionPerClass) - max_hole_size):
                    window = binPredictionPerClass[i: i + max_hole_size]

                    if window[0] == window[-1] == 1:
                        window[:] = [window[0]] * max_hole_size

                # second pass --> split into segments
                converted = []
                cpt = 0
                nb_segment = 0
                previous_elt = None
                for element in binPredictionPerClass:
                    if previous_elt is None:
                        previous_elt = element
                        cpt += 1
                        nb_segment = 1
                        continue

                    if element == previous_elt:
                        cpt += 1

                    else:
                        converted.append((previous_elt, cpt))
                        previous_elt = element
                        nb_segment += 1
                        cpt = 1

                # case where the class is detect during the whole clip
                #                 if nb_segment == 1:
                converted.append((previous_elt, cpt))

                labeled[cls] = converted.copy()
                cls += 1

            output.append(labeled)

        return output

    def _encode_using_median_treshold(self, temporal_prediction: np.array, **kwargs) -> list:
        """ Absolute threshold computed using the median prediction **independently** for each file in the dataset

        Using all the temporal prediction from a specified file, the median of each curve is selected and chosen
        as threshold. Then call the `__encode_using_threshold` function to apply it. The median is retrieve
        independently for each class. Two return mode exist, one that return the average of class's threshold, giving
        an unique threshold for all the class, one that the return the list of threshold, giving a different threshold
        that depend on the class.

        Args:
            temporal_prediction (np.array):
                The complete set for probabilities that need to segmented. must be a three dimensional numpy
                array (<sample>, <class>, <frames>)
            independent (bool):
                If True, return a list of threshold, one different for each class.
                If False, return an unique threshold
            kwargs:
                kwargs ...

        :Exemple:

        ::

            segments = encoder.encode(
                val_strong_pred,
                method="median_threshold",
                **smoothing_parameters  # (optional)
            )

        """
        """ This algorithm is similar to the global median threshold but will compute new threshold(s) (global or \
        independent) for each files.

        Args:
            temporal_prediction (np.array): The complete set for probabilities that need to segmented. must be a \
                three dimensional numpy array (<sample>, <class>, <frames>)
            kwargs:

        Returns:
            The result of the system under the form of a strong annotation text where each line represent one time event.
        """

        # Recover the kwargs arguments
        _global = kwargs.get("global", False)

        output = []

        # Merging "hole" that are smaller than 200 ms
        step_length = self.clip_length / temporal_prediction.shape[1] * 1000
        max_hole_size = int(self.temporal_precision / step_length)

        for clip in temporal_prediction:
            labeled = dict()
            _clip = clip.copy()

            # compute unique threshold for this file
            thresholds = np.array([curve[len(curve) // 2] for curve in clip.T])

            if _global:
                thresholds = [thresholds.mean()] * len(self.classes)

            # Binarize using the given thresholds
            if thresholds is not None:
                _clip[_clip > thresholds] = 1
                _clip[_clip <= thresholds] = 0

            cls = 0
            for binPredictionPerClass in _clip.T:
                # convert the binarized list into a list of tuple representing
                # the element and it's number of # occurrence. The order is
                # conserved and the total sum should be equal to 10s

                # first pass --> Fill the holes
                for i in range(len(binPredictionPerClass) - max_hole_size):
                    window = binPredictionPerClass[i: i + max_hole_size]

                    if window[0] == window[-1] == 1:
                        window[:] = [window[0]] * max_hole_size

                # second pass --> split into segments
                converted = []
                cpt = 0
                nb_segment = 0
                previous_elt = None
                for element in binPredictionPerClass:
                    if previous_elt is None:
                        previous_elt = element
                        cpt += 1
                        nb_segment = 1
                        continue

                    if element == previous_elt:
                        cpt += 1

                    else:
                        converted.append((previous_elt, cpt))
                        previous_elt = element
                        nb_segment += 1
                        cpt = 1

                # case where the class is detect during the whole clip
                #                 if nb_segment == 1:
                converted.append((previous_elt, cpt))

                labeled[cls] = converted.copy()
                cls += 1

            output.append(labeled)

        return output

    def _encode_using_dynamic_threshold(self, temporal_prediction: np.array, **kwargs) -> list:
        raise NotImplementedError()

    # ==========================================================================
    #
    #       SMOOTHING AND UTILITIES
    #
    # ==========================================================================
    def _pad(self, array: np.array, window_size: int,
             method: str = "same") -> np.array:
        """Pad and array using the methods given and a window_size.

        Args:
            array (np.array):
                The array to pad
            window_size (int):
                The size of the working window
            method (str):
                Methods of padding, to be chosen from the following list
                ``` ["same", "valid", "zero"] ```

        Returns:
            the padded array
        """

        output = array

        if method == "same":
            missing = int(window_size / 2)
            first = np.array([array[0]] * missing)
            last = np.array([array[-1]] * missing)

            output = np.concatenate((first, array, last))

        elif method == "valid":
            output = array

        elif method == "zero":
            missing = int(window_size / 2)
            start, end = [0] * missing, [0] * missing

            output = np.concatenate((start, array, end))

        return output

    # ===============================================================================
    #
    #     SMOOTHING FUNCTIONS:
    #
    # ===============================================================================
    def _smooth(self, temporal_prediction: np.array,
                method: str = "smoothMovingAvg",
                **kwargs) -> np.array:
        """ For smoothing the curve of the prediction curves.

        Args:
            temporal_prediction (np.array):
                The temporalPrediction of the second model (TimeDistributed Dense output).
            method (str):
                The algorithm to use for smoothing the curves. ["smoothMovingAvg", "smoothMovingMedian"]
            kwargs:
                See argument list for the smoothing algorithm.
        """

        # Check if methods asked exist
        _methods = ["smoothMovingAvg", "smoothMovingMedian"]
        if method not in _methods:
            raise ValueError("Method %s doesn't exist. Only %s available" %
                             (method, _methods))

        # Create smoother (select the algorithm)
        if method == _methods[0]:
            smoother = self._smooth_moving_avg
        elif method == _methods[1]:
            smoother = self._smooth_moving_median
        else:
            return

        return smoother(temporal_prediction, **kwargs)

    def _smooth_moving_median(self, temporal_prediction: np.array,
                              window_len: int = 11, **kwargs):
        """
        Args:
            temporal_prediction (np.array):
            windows_len (int):
            kwargs:
        """
        raise NotImplementedError()

    def _smooth_moving_avg(self, temporal_prediction: np.array,
                           window_len: int = 5, padding: str = "same",
                           **kwargs):
        """ Apply the smooth moving average on all class. Can be class-dependant or not.

        Args:
            temporal_prediction (np.array):
            window_len (int): The size of the smoothing window, can be
            class-dependant if a list is given
            padding (str): The padding mode to use
            kwargs:
        """

        def smooth(data, _window_len):
            _window_len = int(_window_len)

            if _window_len < 3:
                return data

            s = np.r_[
                2 * data[0] - data[_window_len - 1::-1],
                data,
                2 * data[-1] - data[-1:-_window_len:-1]
            ]

            w = np.ones(_window_len, 'd')
            y = np.convolve(w / w.sum(), s, mode=padding)
            return y[_window_len:-_window_len + 1]

        # core
        smoothed_temporal_prediction = temporal_prediction.copy()

        # In only one window len given, same for all class
        # if nb window_len = nb cls --> one for each class
        windows_len = window_len
        if not isinstance(window_len, Iterable):
            windows_len = [window_len] * len(self.classes)

        for clip_ind in range(len(smoothed_temporal_prediction)):
            clip = smoothed_temporal_prediction[clip_ind]

            for cls_ind in range(len(clip.T)):
                clip.T[cls_ind] = smooth(clip.T[cls_ind], windows_len[cls_ind])

        return smoothed_temporal_prediction




import numpy as np
import optimizers

def square_plus_2(x):
    return - (x**2 + np.sin(x*np.pi))

optimizer = optimizers.DichotomicOptimizer(
    step=100,
    nb_recurse=5,
    nb_process=2,
)

optimizer.fit(square_plus_2, 
    kwds={
        "x": (-2, 2),
    },
)

best_y, best_score = optimizer.best
print("best_y: ", best_y)
print("calc: ", square_plus_2(**best_y))
print("min: ", np.min([square_plus_2(x) for x in np.linspace(-2, 2, 1000)]))
print("max: ", np.max([square_plus_2(x) for x in np.linspace(-2, 2, 1000)]))



import setuptools

long_description = ""

setuptools.setup(
    name="aeseg",
    version="0.0.18",
    author="Leo Cances",
    author_email="leo.cances@gmail.com",
    description="A python package for Sound Event Detection post-processing and optimization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aeseg/aeseg.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
)

